#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Test neutral file generated wiith MYSTRAN
"""

import os
import tempfile

import numpy as np
import pytest
from click.testing import CliRunner

from femap_neutral_parser import Parser, blocks

from .test_utils import datadir, tempdir  # fixtrures


@pytest.fixture(scope="module")
def neutral(datadir):
    fpath = os.path.join(datadir, "mystran_00.NEU")
    return Parser(fpath)


@pytest.fixture(scope="module")
def f_neutral(datadir):
    fpath = os.path.join(datadir, "FEMAP_v8-2.neu")
    return Parser(fpath)


def test_basics(neutral):
    assert neutral.header["femap_version"] == "8.2"
    assert neutral.header["db_title"] == "323152540, bulk_model.F06"
    assert neutral.femap_version == "8.2"
    assert neutral.db_title == "323152540, bulk_model.F06"
    # test table of content
    assert neutral._toc == {
        100: [[3, 5]],
        450: [[8, 13], [907, 912]],
        451: [[16, 904], [915, 1598]],
    }
    # check available blocks
    assert neutral.available_blocks() == {
        "header": 100,
        "output_sets": 450,
        "output_vectors": 451,
    }


# =============================================================================
# Block 450 {{1
# =============================================================================
def test_b450_raw_data(neutral):
    """test Block 450 parsing. Since several Blocks 450 are usually found,
    `_get_block_raw_data` yields a file-like objecy containing each of this
    block
    """
    expected = [
        [
            "       1,\n",
            "Analyse. NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - lc1. test,          \n",
            "0,1,\n",
            "0.,\n",
            "0,\n",
        ],
        [
            "       2,\n",
            "Analyse. NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - lc2. test,     \n",
            "0,1,\n",
            "0.,\n",
            "0,\n",
        ],
    ]

    # =========================================================================
    # test raw parsing
    # =========================================================================
    b450s = neutral._get_block_raw_data(450, as_file_like=False)
    for i, b450 in enumerate(b450s):
        # b450 = b450.readlines()
        assert b450 == expected[i]
    # =========================================================================
    # also test file-like parsing
    # =========================================================================
    b450s = neutral._get_block_raw_data(450, as_file_like=True)
    for i, b450 in enumerate(b450s):
        b450 = b450.readlines()
        assert b450 == expected[i]


def test_b450_data(neutral):
    """test Block 450 parsing.     """
    expected = [
        {
            "ID": 1,
            "anal_type": 1,
            "from_prog": 0,
            "notes": "",
            "title": "Analyse. NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - lc1. test",
            "value": 0.0,
        },
        {
            "ID": 2,
            "anal_type": 1,
            "from_prog": 0,
            "notes": "",
            "title": "Analyse. NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - lc2. test",
            "value": 0.0,
        },
    ]
    parsed_block = neutral._parse_blocks(450)
    assert parsed_block.data == expected


def test_b450_get(neutral):
    """test Block 450 parsing.     """
    expected = {
        1: {
            "anal_type": "Static",
            "from_prog": "Unknown",
            "integer_format": None,
            "notes": "",
            "process_type": None,
            "title": "Analyse. NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - lc1. test",
            "value": 0.0,
        },
        2: {
            "anal_type": "Static",
            "from_prog": "Unknown",
            "integer_format": None,
            "notes": "",
            "process_type": None,
            "title": "Analyse. NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - lc2. test",
            "value": 0.0,
        },
    }

    assert neutral.output_sets == expected


# =============================================================================
# Block 451 {{1
# =============================================================================


def test_b451_data(neutral):
    """test Block 451 parsing.     """
    expected = [
        {
            "ID": 1,
            "anal_type": 1,
            "from_prog": 0,
            "notes": "",
            "title": "Analyse. NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - lc1. test",
            "value": 0.0,
        },
        {
            "ID": 2,
            "anal_type": 1,
            "from_prog": 0,
            "notes": "",
            "title": "Analyse. NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - lc2. test",
            "value": 0.0,
        },
    ]
    parsed_block = neutral._parse_blocks(451)
    data = parsed_block.data
    assert len(data) == 84
    assert set(data[0].keys()) == {
        "abs_max",
        "calc_warn",
        "cent_total",
        "comp_dir",
        "component_vec",
        "ent_type",
        "id_max",
        "id_min",
        "max_val",
        "min_val",
        "out_type",
        "record",
        "setID",
        "title",
        "vecID",
    }

    assert data[0]["vecID"] == 10001
    assert data[1]["vecID"] == 10002
    # ...
    assert data[-1]["vecID"] == 50212
    # -------------------------------------------------------------------------
    # full check for last vector
    expected = {
        "abs_max": 45.04958,
        "calc_warn": True,
        "cent_total": True,
        "comp_dir": 3,
        "component_vec": [50201.0, 50202.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        "ent_type": 8,
        "id_max": 1,
        "id_min": 4,
        "max_val": 45.04958,
        "min_val": -2.221523,
        "out_type": 3,
        "setID": 2,
        "title": "BAR EndB Plane1 Moment",
        "vecID": 50202,
    }
    # we drop the array frm data structure since it requires a specific test
    act_without_array = data[-11].copy()
    actual_records = act_without_array.pop("record")
    assert act_without_array == expected
    expected_records = np.array(
        [
            (1, 4.504958e01),
            (2, 5.241556e-01),
            (3, 1.331676e01),
            (4, -2.221523e00),
            (5, 1.318767e-11),
            (6, -1.419975e-14),
            (9, -2.878763e-01),
            (10, 6.273498e-01),
            (11, 5.241556e-01),
        ],
        dtype=[("entityID", "<i8"), ("value", "<f8")],
    )
    np.testing.assert_array_equal(actual_records, expected_records)


def test_b451_get(neutral, f_neutral):
    """test Block 451 parsing.     """
    expected = {
        1: {
            "anal_type": "Static",
            "from_prog": "Unknown",
            "integer_format": None,
            "notes": "",
            "process_type": None,
            "title": "Analyse. NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - lc1. test",
            "value": 0.0,
        },
        2: {
            "anal_type": "Static",
            "from_prog": "Unknown",
            "integer_format": None,
            "notes": "",
            "process_type": None,
            "title": "Analyse. NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - lc2. test",
            "value": 0.0,
        },
    }
    # -------------------------------------------------------------------------
    # check lazy loading
    assert "output_vectors" not in dir(neutral)
    rss_translation_LCID2 = neutral.output_vectors["Total Translation"][2]
    assert "output_vectors" in dir(neutral)
    values = rss_translation_LCID2.pop("record")
    assert rss_translation_LCID2 == {
        "abs_max": 2.578386,
        "calc_warn": True,
        "cent_total": True,
        "comp_dir": 1,
        "component_vec": [10002.0, 10003.0, 10004.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        "id_max": 7,
        "id_min": 1,
        "max_val": 2.578386,
        "min_val": 0.0,
        "vecID": 1,
    }
    expected_values = np.array(
        [
            (1, 0.000000e00),
            (2, 2.045391e-01),
            (3, 0.000000e00),
            (4, 1.468270e-02),
            (5, 9.231050e-05),
            (6, 6.276400e-01),
            (7, 2.578386e00),
            (8, 1.025100e-01),
            (9, 2.578363e00),
            (10, 1.916094e00),
            (11, 1.100510e00),
            (12, 2.389742e00),
        ],
        dtype=[("nodeID", "<i8"), ("disp", "<f8")],
    )
    np.testing.assert_array_equal(values, expected_values)



"""
This module contains defined blocks. Any block shall inherit from `Block`.
"""

import logging
from collections import defaultdict

from femap_neutral_parser.blocks._base import Block

# =============================================================================
# FEMAP : (FEMAP vectorID, MYSTRAN title, MYSTRAN vectorID)
# =============================================================================
MYSTRAN2FEMAP = {
    "RSS translation": ("Total Translation", 1),
    "T1  translation": ("T1 Translation", 2),
    "T2  translation": ("T2 Translation", 3),
    "T3  translation": ("T3 Translation", 4),
    "RSS rotation": ("Total Rotation", 5),
    "R1  rotation": ("R1 Rotation", 6),
    "R2  rotation": ("R2 Rotation", 7),
    "R3  rotation": ("R3 Rotation", 8),
    "RSS applied force": ("Total Applied Force", 41),
    "T1  applied force": ("T1 Applied Force", 42),
    "T2  applied force": ("T2 Applied Force", 43),
    "T3  applied force": ("T3 Applied Force", 44),
    "RSS applied moment": ("Total Applied Moment", 45),
    "R1  applied moment": ("R1 Applied Moment", 46),
    "R2  applied moment": ("R2 Applied Moment", 47),
    "R3  applied moment": ("R3 Applied Moment", 48),
    "RSS SPC force": ("Total Constraint Force", 51),
    "T1  SPC force": ("T1 Constraint Force", 52),
    "T2  SPC force": ("T2 Constraint Force", 53),
    "T3  SPC force": ("T3 Constraint Force", 54),
    "RSS SPC moment": ("Total Constraint Moment", 55),
    "R1  SPC moment": ("R1 Constraint Moment", 56),
    "R2  SPC moment": ("R2 Constraint Moment", 57),
    "R3  SPC moment": ("R3 Constraint Moment", 58),
    None: ("Elem C2 R3 GPMoment", 85015),
    "BAR EndA Plane1 Moment": ("Bar EndA Plane1 Moment", 3000),
    "BAR EndA Plane2 Moment": ("Bar EndA Plane2 Moment", 3001),
    "BAR EndB Plane1 Moment": ("Bar EndB Plane1 Moment", 3002),
    "BAR EndB Plane2 Moment": ("Bar EndB Plane2 Moment", 3003),
    "BAR EndA Pl1 Shear Force": ("Bar EndA Pl1 Shear Force", 3004),
    "BAR EndA Pl2 Shear Force": ("Bar EndA Pl2 Shear Force", 3005),
    "BAR EndA Axial Force": ("Bar EndA Axial Force", 3008),
    "BAR EndA Torque": ("Bar EndA Torque", 3010),
    "BAR EndA Pt1 Comb Stress": ("Bar EndA Pt1 Bend Stress", 3075),
    "BAR EndA Pt2 Comb Stress": ("Bar EndA Pt2 Bend Stress", 3076),
    "BAR EndA Pt3 Comb Stress": ("Bar EndA Pt3 Bend Stress", 3077),
    "BAR EndA Pt4 Comb Stress": ("Bar EndA Pt4 Bend Stress", 3078),
    "BAR EndB Pt1 Comb Stress": ("Bar EndB Pt1 Bend Stress", 3083),
    "BAR EndB Pt2 Comb Stress": ("Bar EndB Pt2 Bend Stress", 3084),
    "BAR EndB Pt3 Comb Stress": ("Bar EndB Pt3 Bend Stress", 3085),
    "BAR EndB Pt4 Comb Stress": ("Bar EndB Pt4 Bend Stress", 3086),
    "BAR EndA Max Stress": ("Bar EndA Max Comb Stress", 3109),
    "BAR EndA Min Stress": ("Bar EndA Min Comb Stress", 3110),
    "BAR EndB Max Stress": ("Bar EndB Max Comb Stress", 3111),
    "BAR EndB Min Stress": ("Bar EndB Min Comb Stress", 3112),
}
# =============================================================================
# Block B451
# =============================================================================


class B451(Block):
    """
    >>> txt = '''       1,   10001,       1
    ... RSS translation
    ...      0.000000E+00,     1.602912E+00,     1.602912E+00,
    ...    10002,   10003,   10004,       0,       0,       0,       0,       0,       0,       0,
    ...        0,       0,       0,       0,       0,       0,       0,       0,       0,       0,
    ...        1,      12,       1,       7,
    ...        1,       1,       1
    ...        1,     0.000000E+00,
    ...        2,     1.870816E-01,
    ...        3,     0.000000E+00,
    ...        4,     7.011176E-02,
    ...       -1,     0.          ,
    ... '''
    >>> b = B451(version=9.3)
    >>> b.parse(txt)
    >>> from pprint import pprint as pp
    >>> import pprint; pprint.pprint(b.data)
    [{'abs_max': 1.602912,
     'calc_warn': True,
     'cent_total': True,
     'comp_dir': 1,
     'component_vec': [10002.0, 10003.0, 10004.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
     'ent_type': 7,
     'id_max': 12,
     'id_min': 1,
     'max_val': 1.602912,
     'min_val': 0.0,
     'out_type': 1,
     'record': array([(1, 0.        ), (2, 0.1870816 ), (3, 0.        ), (4, 0.07011176)],
          dtype=[('entityID', '<i8'), ('value', '<f8')]),
     'setID': 1,
     'title': 'RSS translation',
     'vecID': 10001}]
    """

    NAME = "output_vectors"
    RECORDS = [
        [
            {"field": "setID", "coerce": int},
            {"field": "vecID", "coerce": int},
            {"field": "_", "coerce": int},
        ],  # 1
        [
            {"field": "title"},
        ],  # 2
        [
            {"field": "min_val", "coerce": float},  # 3
            {"field": "max_val", "coerce": float},
            {"field": "abs_max", "coerce": float},
        ],
        [  # 4 comp[0..9]
            {"field": "component_vec", "coerce": float, "single_line_array": 9},
        ],
        [  # 5 comp[10..19]
            {"field": "_", "coerce": float, "single_line_array": 9},
        ],
        [  # 6
            {"field": "doubled_sided_contour", "coerce": int, "v+": "10.0"},
        ],
        [  # 7
            {"field": "id_min", "coerce": int},
            {"field": "id_max", "coerce": int},
            {"field": "out_type", "coerce": int},
            {"field": "ent_type", "coerce": int},
        ],
        [  # 8
            {"field": "calc_warn", "coerce": bool},
            {"field": "comp_dir", "coerce": int, "v+": "4.1"},
            {"field": "cent_total", "coerce": bool},
        ],
        [  # 1 record for each entry, plus last record
            {
                "multi_line_array": True,
                "field": "record",
                "fields": ("entityID", "value"),
                "coerce": (int, float),
            },
        ],
    ]

    def digest(self):
        agg = defaultdict(dict)
        for data in self.data:
            # -----------------------------------------------------------------
            # affect appropriate columns headers
            ent_type = {7: "nodeID", 8: "eltID"}[data.pop("ent_type")]
            out_type = {
                0: "value",
                1: "disp",
                2: "accel",
                3: "force",
                4: "stress",
                5: "strain",
                6: "temp",
            }.get(data.pop("out_type"), "user")
            data["record"].dtype.names = (ent_type, out_type)
            # -----------------------------------------------------------------
            # modify MYSTRAN titles and setIDs
            title = data.pop("title")
            title, vecID = MYSTRAN2FEMAP.get(title, (title, data["vecID"]))
            data["vecID"] = vecID
            agg[title][data.pop("setID")] = data
        return dict(agg)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)
    # import sys
    # filepath =sys.argv[1]



import os

import pytest


@pytest.fixture
def tempdir():
    tempdir = tempfile.mkdtemp()
    return tempdir


@pytest.fixture(scope="module")
def datadir(request):
    """
    Fixture responsible for searching a folder with the same name of test
    module and, if available, moving all contents to a temporary directory so
    tests can use them freely.
    """
    filename = request.module.__file__
    basename, _ = os.path.split(filename)
    return os.path.join(basename, "data")



import logging
import os
from collections import defaultdict as dd
from io import StringIO
from typing import Dict, List, Tuple

import numpy as np

from femap_neutral_parser import blocks


class Parser:
    def __init__(self, fpath: str) -> None:
        self.fpath = os.path.expanduser(fpath)
        self._toc = dd(list)
        if not os.path.isfile(self.fpath):
            raise ValueError(f'file path "{fpath}" not found')
        self._build_toc()
        self._parse_header()
        # self.blocks = {}
        # ---------------------------------------------------------------------
        # collect available defined blocks (code-wise, **not** neutral file wise)
        self._defined_block_names = {
            getattr(blocks, b).NAME: getattr(blocks, b)
            for b in dir(blocks)
            if b.startswith("B")
        }

    def available_blocks(self) -> Dict[int, str]:
        return {
            getattr(blocks, b).NAME: getattr(blocks, b).id()
            for b in dir(blocks)
            if b.startswith("B")
        }

    def __getattr__(self, key: str):
        # look for blocks ahving key as NAME
        block = self._parse_blocks(self._defined_block_names[key].id())
        data = block.digest()
        setattr(self, key, data)
        return data

    def _parse_header(self):
        """parse block 100. This needs to be done at low-level since we will
        pick FEMAP's version number from there."""
        txt = tuple(self._get_block_raw_data(100))[0]
        b100 = blocks.B100(version=0)  # we do not have correct version at this point
        b100.parse(txt)
        self.header = b100.data[0]
        # also assign directly for convenieance
        self.femap_version = b100.data[0]["femap_version"]
        self.db_title = b100.data[0]["db_title"]

    def _build_toc(self):
        """parse the document to retrieve blocks locations"""
        nb_lines = 0
        current_block = None
        for line_nb, line in enumerate(open(self.fpath, "r")):
            # if line_nb == 1598:
            #     breakpoint()
            line = line.strip()
            if line == "-1":
                if not current_block:
                    # ---------------------------------------------------------
                    # new block is coming, block ID known the line after
                    # ---------------------------------------------------------
                    current_block = -1  # set waiting status
                    continue
                else:
                    # ---------------------------------------------------------
                    # block is finishing
                    # ---------------------------------------------------------
                    self._toc[current_block][-1].append(line_nb)
                    current_block = None
                    continue
            if current_block == -1:
                # -------------------------------------------------------------
                # waiting status
                current_block = int(line)
                self._toc[current_block].append([line_nb + 1])
            nb_lines += 1
        # ---------------------------------------------------------------------
        # EOF
        # we sometimes miss trailing "-1". fix it:
        if current_block is not None and current_block > 0:
            self._toc[current_block][-1].append(line_nb)
        self._toc = dict(self._toc)
        logging.info("parsed %d lines", nb_lines)

    def _get_block_raw_data(self, block_id, as_file_like=False):
        """yield file-like objects containing block data"""
        fh = open(self.fpath)
        for line_start, line_end in self._toc[block_id]:
            fh.seek(0)
            lines = range(line_start, line_end)
            lines_content = []
            for line_nb, line in enumerate(fh):
                if line_nb > line_end:
                    break
                if line_nb in lines:
                    lines_content.append(line)  # .replace("<NULL>", ""))
            if as_file_like:
                txt = StringIO()
                txt.writelines(lines_content)
                txt.seek(0)
            else:
                txt = lines_content
            yield txt
        fh.close()

    def _parse_blocks(self, block_id: int) -> None:
        """"""
        block = getattr(blocks, f"B{block_id}")(version=self.femap_version)
        for txt in self._get_block_raw_data(block_id=block_id):
            block.parse(txt)
        return block


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

with open("requirements_dev.txt") as f:
    test_requirements = f.read().splitlines()

setup_requirements = [
    "pytest-runner",
]

setup(
    author="Nicolas Cordier",
    author_email="nicolas.cordier@numeric-gmbh.ch",
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    description="FEMAP neutral file parser",
    entry_points={
        "console_scripts": [
            "femap_neutral_parser=femap_neutral_parser.cli:main",
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="femap_neutral_parser",
    name="femap_neutral_parser",
    packages=find_packages(include=["femap_neutral_parser", "femap_neutral_parser.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    version="0.1.0",
    zip_safe=False,
    url="https://framagit.org/numenic/femap_neutral_parser",
)
